/******************************************************************************
 * CoreWall / Corelyzer - An Initial Core Description Tool
 * Copyright (C) 2005 Arun Rao, Electronic Visualization Laboratory,  
 * University of Illinois at Chicago
 *
 * This software is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either Version 2.1 of the License, or 
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public 
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser Public License along
 * with this software; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "baseimghandler.h"
#include "utility.h"

#define DEBUG_IMG_HANDLER 1

DLL_EXPORT ImageHandler::ImageHandler()
{
    width = 0;
    height = 0;
    pixels = NULL;
    resH = 0;
    pixelFormat = UNDEFINED_FORMAT;
    resFormat = UNDEFINED_FORMAT;
	
    sourceFile = NULL;
}

DLL_EXPORT ImageHandler::ImageHandler(const ImageHandler& other, float scale )
{

	pixelFormat = other.pixelFormat;
	resFormat = other.resFormat;
	resH = other.resH * scale;
	resV = other.resV * scale;
	width = int(other.width * scale);
	height = int(other.height * scale);
	if( other.sourceFile )
	{
		sourceFile = new char[ strlen( other.sourceFile ) + 1];
		strcpy( sourceFile, other.sourceFile );
	}

	int components = 3;
	if( other.pixelFormat == RGBA || other.pixelFormat == BGRA )
		components = 4;

	if( scale != 1.0 )
	{
		this->pixels = NULL;
		
		ResampleBuffer( other.pixels, other.width, other.height, 
			components, scale, pixels, width, height, false );	
	}
	else
	{
		memcpy( pixels, other.pixels, width * height * components );
	}

}

DLL_EXPORT ImageHandler::~ImageHandler()
{
    if( pixels )
        delete [] pixels;

    if( sourceFile )
	delete [] sourceFile;

}


DLL_EXPORT bool ImageHandler::Load(char* filename)
{
	return false;
}

DLL_EXPORT char* ImageHandler::GetPixels()
{
    return pixels;
}

DLL_EXPORT int ImageHandler::GetPixelFormat()
{
    return pixelFormat;
}

DLL_EXPORT int ImageHandler::GetWidth()
{
    return width;
}

DLL_EXPORT int ImageHandler::GetHeight()
{
    return height;
}

DLL_EXPORT float ImageHandler::GetResolutionH()
{
    return resH;
}

DLL_EXPORT float ImageHandler::GetResolutionV()
{
	return resV;
}

DLL_EXPORT int ImageHandler::GetResolutionFormat()
{
    return resFormat;
}

DLL_EXPORT void ImageHandler::CloseImage()
{
    if( pixels )
        delete [] pixels;
        
    pixels = NULL;
    width = 0;
    height = 0;
    resH = 0;
    pixelFormat = UNDEFINED_FORMAT;
    resFormat = UNDEFINED_FORMAT;
}

DLL_EXPORT char* ImageHandler::GetSubImage( char* outbuf, 
		int x, int y, int w, int h, float scale )
{
//	printf("Called get sub image: %d, %d for %d x %d at scale %f\n",
//		x,y,w,h,scale);
	if( scale == 0)
	{
		printf("Scale = 0! returning\n");
		return NULL;
	}

	if( x < 0 || x >= width )
	{	
		printf("Start x not in picture bounds. pic width %d!\n", width);
		return NULL;
	}
	
	if( y < 0 || y >= height )
	{
		printf("Start y not in picture bounds! pic height %d!\n", height);
		return NULL;
	}

	char* copy;
	int components = 3;
	if( pixelFormat == RGBA || pixelFormat == BGRA )
		components = 4;
	
	int copyW, copyH;
	copyW = int(w * scale);
	copyH = int(h * scale);
	int skipX, skipY;
	skipX = copyW / w;
	skipY = copyH / h;
//	printf("Starting at %d, %d Skipping %d x pixels, %d y pixels\n", 
//		x, y, skipX, skipY);

	copy = outbuf;

	memset( copy, 0, w * h * components );
	//copy the sub image over
	if( width < x + w )
		copyW = width - x;
	if( height < y + h )
		copyH = height - y;
	
	for( int i = 0; i < h; i++)
	{
		for( int j = 0; j < w; j++)
		{
			for( int k = 0; k < components; k++)
			{
				if( (y + (i * skipY)) < height &&
					(x + (j * skipX)) < width )
				{
					copy[ (i * w + j) * components + k] =
						pixels[ ((y + (i * skipY)) * width +
							(x + (j * skipX))) * components + k];
				}
				else
				{
					copy[ (i * w + j) * components + k] = 0;
				}
			}	
		}
	}

	return outbuf;
}


DLL_EXPORT void ImageHandler::FreeSubImage(char* subimg)
{
	if( subimg )
		delete [] subimg;
}

DLL_EXPORT const char* ImageHandler::GetSourceFilename()
{
	return sourceFile;
}

DLL_EXPORT void ImageHandler::Rotate90DegCW()
{
    float tempfloat;
    int tempint;
	int components = 3;
	if( pixelFormat == RGBA || pixelFormat == BGRA )
		components = 4;
    
    tempint = width;
    width = height;
    height = tempint;

    tempfloat = resH;
    resH = resV;
    resV = tempfloat;
   
    char* newpixels = new char[ width * height * components ];

    for( int i = 0; i < width; i++)
    {
        for( int k = 0; k < height; k++)
        {
            int rr, rc;
            rr = width - i;
            rc = k;
            for(int c = 0; c < components; c++)
            {
                newpixels[k * width * components + i * components + c] =
                    pixels[rr * height * components + rc * components + c]; 
            } // for each component
        } // for each row
    } //for each column

    delete [] pixels;
    pixels = newpixels;
}


DLL_EXPORT void ImageHandler::Rotate90DegCCW()
{
    float tempfloat;
    int tempint;
	int components = 3;
	if( pixelFormat == RGBA || pixelFormat == BGRA )
		components = 4;
    
    tempint = width;
    width = height;
    height = tempint;

    tempfloat = resH;
    resH = resV;
    resV = tempfloat;
   
    char* newpixels = new char[ width * height * components ];

    for( int i = 0; i < width; i++)
    {
        for( int k = 0; k < height; k++)
        {
            int rr, rc;
            rr = i;
            rc = height - k;
            for(int c = 0; c < components; c++)
            {
                newpixels[k * width * components + i * components + c] =
                    pixels[rr * height * components + rc * components + c]; 
            } // for each component
        } // for each row
    } //for each column

    delete [] pixels;
    pixels = newpixels;
}


DLL_EXPORT void ImageHandler::Rotate180Deg()
{
    Rotate90DegCW();
    Rotate90DegCW();
}


