#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <unistd.h>

#include "cpgl.h"
#include "trimodel.h"
#include "jpegimghandler.h"
#include "model.h"
#include "obj.h"

#ifndef PI
#define PI 3.14159
#endif

#define DEG2RAD( x ) ( (x) / 180.0f * PI )
#define COLORMAP 0
#define NORMALS  1
#define DEPTHS   2
#define NDMAP    3

extern "C" {
    extern void snap(char* filename, int w, int h);
}

//-----------------------------------
SDL_Surface* gDrawSurface = NULL;
int width  = 512;
int height = 512;

//-----------------------------------
// Update Function Pointer and Variables Updated Each Call
void (*updatefn)(double) = NULL;
float    TotalTime   = 0;
GLfloat  LightPos[4] = { 8, 8, 8, 0};
GLfloat  LightVel[3] = { 0, 0, 0 };
GLfloat  CamMat[16]  = { 1, 0, 0, 0,
                         0, 1, 0, 0,
                         0, 0, 1, 0,
                         0, 0, 0, 1 };
GLfloat  CamPos[3]   = { 0, 0, 0 };
GLfloat  CamVel[3]   = { 0, 0, 0 };
bool     BreakLoop   = false;
bool     Capture     = false;
char*    MapFileName = NULL;
char*    ColorFileName = NULL;
GLubyte* Pixels      = NULL;
int      DrawStyle   = COLORMAP;

//----------------------------------
// obj handles
int reliefObj = -1;

//----------------------------------

//----------------------------------
// App specific functions

//-----------------------------------
// Function prototypes
char* get_shader_source(const char* filename);
void setup_shaders(const char* vert, const char* frag,GLhandleARB& vs,
                   GLhandleARB& fs, GLhandleARB& prg);
void unload_shaders(GLhandleARB& vs, GLhandleARB& fs, GLhandleARB& prg);
void draw_scene();
void proc_key_down(SDL_Event* e);
void proc_key_up(SDL_Event* e);
void inf_loop();
void update(double dt);
void init_light();
void setup_sdl();
void setup_gl_procs();

#if !defined(__APPLE__) && !defined(_WIN32)

// SHADER FUNCTIONS
PFNGLCREATEPROGRAMOBJECTARBPROC     glCreateProgramObjectARB = NULL;
PFNGLCREATESHADEROBJECTARBPROC      glCreateShaderObjectARB = NULL;
PFNGLSHADERSOURCEARBPROC            glShaderSourceARB = NULL;
PFNGLCOMPILESHADERARBPROC           glCompileShaderARB = NULL;
PFNGLGETOBJECTPARAMETERIVARBPROC    glGetObjectParameterivARB = NULL;
PFNGLATTACHOBJECTARBPROC            glAttachObjectARB = NULL;
PFNGLGETINFOLOGARBPROC              glGetInfoLogARB = NULL;
PFNGLLINKPROGRAMARBPROC             glLinkProgramARB = NULL;
PFNGLUSEPROGRAMOBJECTARBPROC        glUseProgramObjectARB = NULL;
PFNGLGETUNIFORMLOCATIONARBPROC      glGetUniformLocationARB = NULL;
PFNGLUNIFORM1FARBPROC               glUniform1f = NULL;
PFNGLUNIFORM2FARBPROC               glUniform2f = NULL;
PFNGLUNIFORM3FARBPROC               glUniform3f = NULL;
PFNGLUNIFORM1FVARBPROC              glUniform1fv = NULL;
PFNGLUNIFORM2FVARBPROC              glUniform2fv = NULL;
PFNGLUNIFORM3FVARBPROC              glUniform3fv = NULL;
PFNGLUNIFORM1IPROC                  glUniform1i = NULL;

// MULTITEXTURE FUNCTIONS
PFNGLACTIVETEXTUREPROC              glActiveTexture = NULL;

#endif

//-----------------------------------
int main(int argc, char** argv)
{
    if( argc < 4 ) return 0;

    setup_sdl();
    setup_gl_procs();
    init_light();
    updatefn = update;
    obj_add_file(NULL);
    // try to load relief object
    reliefObj = load_model(argv[1],
                           "shaders/normalmap.vert",
                           "shaders/normalmap.frag"); 
    MapFileName = new char[ strlen(argv[2]) ];
    strcpy(MapFileName,argv[2]);
    ColorFileName = new char[ strlen(argv[3]) ];
    strcpy(ColorFileName, argv[3]);

    Pixels = new GLubyte[ width * height * 4 ];

    inf_loop();
    free_all_models();
    obj_del_file(0);

    delete [] Pixels;
    delete [] MapFileName;
    delete [] ColorFileName;

    return 0;
}

//-----------------------------------
// Load Shader Source From File
char* get_shader_source(const char* filename)
{
    char* source;
    FILE* fptr = fopen(filename,"rb");
    if( !fptr )
        return NULL;
    fseek(fptr,0,SEEK_END);
    long length = ftell(fptr);
    source = new char[length + 1];
    memset(source,0,length + 1);
    fseek(fptr,0,SEEK_SET);
    fread(source,sizeof(char),length,fptr);
    fclose(fptr);

    return source;
}

//-----------------------------------
// Setup shaders
void setup_shaders(const char* vert,const char* frag, GLhandleARB& vs, 
                   GLhandleARB& fs, GLhandleARB& prg)
{
    bool vfailed = false;
    bool ffailed = false;

    char infolog[1024];
    GLint length;
    // get shader source code
    char* vertsource = get_shader_source(vert);
    char* fragsource = get_shader_source(frag);
    if(!vertsource && !fragsource )
    {
        fprintf(stderr,"No shader source code!\n");
        return;
    }

    prg = glCreateProgramObjectARB();
    fs = 0;
    vs = 0;
    // process the vertex shader source
    if( vertsource )
    {    
        vs = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
        glShaderSourceARB(vs,1,(const GLcharARB**)&vertsource,NULL);
        glCompileShaderARB(vs);
       
        // see if it compiled
        GLint compiled;
        glGetObjectParameterivARB(vs,GL_OBJECT_COMPILE_STATUS_ARB,
            &compiled);
        
        if(compiled)
            glAttachObjectARB(prg,vs);
        else
        {
            printf("VERTEX SHADER COMPILE PROBLEM\n");
            glGetInfoLogARB(vs,1024,&length,infolog);
            if( length ) printf("GLSL PROGRAM LOG: %s\n", infolog);
        }
    }
   
    // process the fragment shader source code
    if( fragsource )
    {
        fs = glCreateShaderObjectARB(GL_FRAGMENT_SHADER);
        glShaderSourceARB(fs,1,(const GLcharARB**)&fragsource,NULL);
        glCompileShaderARB(fs);
       
        // see if the fragment shader compiled
        GLint compiled;
        glGetObjectParameterivARB(fs,GL_OBJECT_COMPILE_STATUS_ARB,
            &compiled);
        
        if(compiled)
            glAttachObjectARB(prg,fs);
        else
        {
            printf("FRAG SHADER COMPILE PROBLEM\n");
            glGetInfoLogARB(fs,1024,&length,infolog);
            if( length ) printf("GLSL PROGRAM LOG: %s\n", infolog);
        }
    }

    glLinkProgramARB(prg);
    
    glGetInfoLogARB(prg,1024,&length,infolog);
    if( length )
    {
        printf("GLSL PROGRAM LOG: %s\n", infolog );
    }
    
    glUseProgramObjectARB(prg);


    if( fragsource ) delete [] fragsource;
    if( vertsource ) delete [] vertsource;

}

//-----------------------------------
// unload shaders
void unload_shaders(GLhandleARB& vs, GLhandleARB& fs, GLhandleARB& prg)
{
    glDeleteObjectARB(prg);
    vs = 0;
    fs = 0;
    prg = 0;
}

//-----------------------------------
// Draw
void draw_scene()
{
    glClearColor(0, 0, 0, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    glEnable(GL_CULL_FACE);
    glDisable(GL_BLEND);
    glCullFace(GL_BACK);
    glEnable(GL_DEPTH_TEST);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-1,1,-1,1,0,1);
    glMatrixMode(GL_MODELVIEW);
    
    draw_model(reliefObj);

    SDL_GL_SwapBuffers();
    
    if( Capture )
    {
        /*Capture = false;
        glReadPixels(0,0,width,height,GL_RGBA, GL_UNSIGNED_BYTE, Pixels);
       
        FILE* fptr;
        if( DrawStyle == COLORMAP )
            fptr = fopen(ColorFileName,"wb");
        else
            fptr = fopen(MapFileName,"wb");

        if( !fptr )
            printf("Can't open to write out\n");
        else
        {
            fprintf( fptr, "P6\n%d %d\n255\n", width, height);
            fwrite( Pixels, sizeof(GLubyte), width * height * 4, fptr);
            fclose(fptr);
        }
        */ 
        Capture = false;
        if(DrawStyle == COLORMAP) snap(ColorFileName,width,height);
        else snap(MapFileName,width,height);
    }
}

//-----------------------------------
// Process key up presses
void proc_key_up(SDL_Event* e)
{
    if( (e->key.keysym.unicode & 0xFF80) == 0)
    {
        // it's an ascii character
        char ch = e->key.keysym.unicode & 0x7F;
        switch( ch )
        {
        case 'i':
        case 'k':
            CamVel[0] = CamVel[1] = CamVel[2] = 0;
            break;
        }
    }

    switch( e->key.keysym.sym )
    {
    case SDLK_LEFT:
    case SDLK_RIGHT:
    case SDLK_UP:
    case SDLK_DOWN:
        LightVel[0] = LightVel[1] = LightVel[2] = 0;
        break;
    }
}

//-----------------------------------
// Process key down presses
void proc_key_down(SDL_Event* e)
{
    if( (e->key.keysym.unicode & 0xFF80) == 0)
    {
        // it's an ascii character
        char ch = e->key.keysym.unicode & 0x7F;
        switch( ch )
        {
        case 'i':
            CamVel[2] -= 10;
            break;
        case 'k':
            CamVel[2] += 10;
            break;
        case '<':
            stretch_model(reliefObj,.9);
            break;
        case '>':
            stretch_model(reliefObj,1.1);
            break;
        case '[':
            turn_model(reliefObj, -.1, 0, 0);
            break;
        case ']':
            turn_model(reliefObj, .1, 0, 0);
            break;
        case 'j':
            translate_model(reliefObj, -.1, 0, 0);
            break;
        case 'l':
            translate_model(reliefObj, .1, 0, 0);
            break;
        case 'c':
        case 'C':
            Capture = true;
            break;
        case '1':
            glUniform1f( get_model_uniform_loc(reliefObj,"Style"), COLORMAP);
            DrawStyle = COLORMAP;
            break;
        case '2':
            glUniform1f( get_model_uniform_loc(reliefObj,"Style"), NORMALS);
            DrawStyle = NORMALS;
            break;
        case '3':
            glUniform1f( get_model_uniform_loc(reliefObj,"Style"), DEPTHS);
            DrawStyle = DEPTHS;
            break;
        case '4':
            glUniform1f( get_model_uniform_loc(reliefObj,"Style"), NDMAP);
            DrawStyle = NDMAP;
            break;
        }
    }

    switch( e->key.keysym.sym )
    {
    case SDLK_ESCAPE:
        BreakLoop = true;
        break;
    case SDLK_LEFT:
        break;
    case SDLK_RIGHT:
        break;
    case SDLK_UP:
        break;
    case SDLK_DOWN:
        break;
    }
}

//-----------------------------------
// Infinite Loop
void inf_loop()
{
    double starttime, endtime;
    starttime = clock();
    endtime = starttime;
    int dt = 0;
    starttime = (double) SDL_GetTicks() / 1000.0; 
    
    //--- infinite loop with event queue processing
    SDL_Event event;
    while(!BreakLoop) 
    {
        while( SDL_PollEvent( &event )) 
        {
            switch( event.type )
            {
            case SDL_KEYDOWN:
                proc_key_down(&event);
                break;
            case SDL_KEYUP:
                proc_key_up(&event);
                break;
            case SDL_QUIT:
                BreakLoop = true;
                break;
            }

        } // -- while event in queue

        if(!BreakLoop)
        {
            //capture frame time
            endtime = (double) SDL_GetTicks() / 1000.0; 
        
            if( endtime - starttime > 1.0f / float(DESIRED_FPS) )
            {
                if( updatefn )
                    updatefn(endtime - starttime);
                starttime = (double) SDL_GetTicks() / 1000.0;
            
                // update scene 
                draw_scene();
            }
        }
        
    } // -- infinite loop
}

//-----------------------------------
// homebrew update funciton
void update(double dt)
{
    TotalTime += dt;
    int i;
    for( i = 0; i < 3; i++)
        CamPos[i] += CamVel[i] * dt;

}

//-----------------------------------
// Setup lights
void init_light()
{
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT1);
    
    GLfloat lightamb[] = { .3, .3, .3, 1 };
    GLfloat lightdif[] = { 1, 1, 1, 1 };
    glShadeModel(GL_SMOOTH);
    glLightfv(GL_LIGHT1,GL_AMBIENT,lightamb);
    glLightfv(GL_LIGHT1,GL_DIFFUSE,lightdif);
    glLightfv(GL_LIGHT1,GL_POSITION,LightPos);

}


//-----------------------------------
// SETUP SDL
void setup_sdl()
{
    // init video system
    const SDL_VideoInfo* vidinfo;
    if( SDL_Init(SDL_INIT_VIDEO) < 0 )
    {
        fprintf(stderr,"Failed to initialize SDL Video!\n");
        exit(1);
    }

    // use unicode characters
    SDL_EnableUNICODE(0);

    // tell system which funciton to process when exit() call is made
    atexit(SDL_Quit);

    // get optimal video settings
    vidinfo = SDL_GetVideoInfo();
    if(!vidinfo)
    {
        fprintf(stderr,"Coudn't get video information!\n%s\n", SDL_GetError());
        exit(1);
    }

    // set opengl attributes
    SDL_GL_SetAttribute(SDL_GL_RED_SIZE,        5);
    SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE,      5);
    SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,       5);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,      32);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,    1);

    // get a framebuffer
    gDrawSurface = SDL_SetVideoMode(width,height,vidinfo->vfmt->BitsPerPixel,
        SDL_OPENGL);

    if( !gDrawSurface )
    {
        fprintf(stderr,"Couldn't set video mode!\n%s\n", SDL_GetError());
        fprintf(stderr,"Trying lower depth size to 24\n");
        SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,      24);
        gDrawSurface = SDL_SetVideoMode(width,height,
            vidinfo->vfmt->BitsPerPixel, SDL_OPENGL);
        if(!gDrawSurface)
        {
            fprintf(stderr,"Couldn't set video mode!\n%s\n", SDL_GetError());
            fprintf(stderr,"Trying lower depth size 16\n");
            SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,      16);
            gDrawSurface = SDL_SetVideoMode(width,height,
                vidinfo->vfmt->BitsPerPixel, SDL_OPENGL);
            if(!gDrawSurface)
            {
                fprintf(stderr,"Couldn't set video mode!\n%s\n", 
                    SDL_GetError());
                exit(1);
            }
        }
    }

    // give the window a name
    SDL_WM_SetCaption("ARUN RAO - CS594 PROJ1", "CS594 PROJ1");
    // set opengl viewport and perspective view
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glViewport(0,0,width,height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective( 60, 4.0f / 3.0f, .01, 10000);
    glMatrixMode(GL_MODELVIEW);
}

#ifdef __APPLE__
void setup_gl_procs()
{
}

#elif !defined(_WIN32)

void setup_gl_procs()
{
    // SHADER PROCS
    glCreateProgramObjectARB = (PFNGLCREATEPROGRAMOBJECTARBPROC)
        glGetProcAddress("glCreateProgramObjectARB");
    glCreateShaderObjectARB = (PFNGLCREATESHADEROBJECTARBPROC)
        glGetProcAddress("glCreateShaderObjectARB");
    glShaderSourceARB = (PFNGLSHADERSOURCEARBPROC)
        glGetProcAddress("glShaderSourceARB");
    glCompileShaderARB = (PFNGLCOMPILESHADERARBPROC)
        glGetProcAddress("glCompileShaderARB");
    glGetObjectParameterivARB = (PFNGLGETOBJECTPARAMETERIVARBPROC)
        glGetProcAddress("glGetObjectParameterivARB");
    glAttachObjectARB = (PFNGLATTACHOBJECTARBPROC)
        glGetProcAddress("glAttachObjectARB");
    glGetInfoLogARB = (PFNGLGETINFOLOGARBPROC)
        glGetProcAddress("glGetInfoLogARB");
    glLinkProgramARB = (PFNGLLINKPROGRAMARBPROC)
        glGetProcAddress("glLinkProgramARB");
    glUseProgramObjectARB = (PFNGLUSEPROGRAMOBJECTARBPROC)
        glGetProcAddress("glUseProgramObjectARB");
    glGetUniformLocationARB = (PFNGLGETUNIFORMLOCATIONARBPROC)
        glGetProcAddress("glGetUniformLocationARB");
    glUniform1f = (PFNGLUNIFORM1FARBPROC)
        glGetProcAddress("glUniform1fARB");
    glUniform2f = (PFNGLUNIFORM2FARBPROC)
        glGetProcAddress("glUniform2fARB");
    glUniform3f = (PFNGLUNIFORM3FARBPROC)
        glGetProcAddress("glUniform3f"); 
    glUniform1fv = (PFNGLUNIFORM1FVARBPROC)
        glGetProcAddress("glUniform1fvARB");
    glUniform2fv = (PFNGLUNIFORM2FVARBPROC)
        glGetProcAddress("glUniform2fvARB");
    glUniform3fv = (PFNGLUNIFORM3FVARBPROC)
        glGetProcAddress("glUniform3fvARB"); 
    glUniform1i = (PFNGLUNIFORM1IPROC)
        glGetProcAddress("glUniform1i");

    // MULTITEXTURE FUNCTIONS
    glActiveTexture = (PFNGLACTIVETEXTUREPROC)
        glGetProcAddress("glActiveTexture");
}

#endif
