#include "model.h"
#include "obj.h"
#include "matrix.h"
#include <vector>
#include <math.h>

//------------------------------------
std::vector< ObjModel* > gModelVec;

// identity matrix in row-major is same as in column-major... :)
const GLfloat identity[16] = { 1.0f, 0.0f, 0.0f, 0.0f,   // first column
                               0.0f, 1.0f, 0.0f, 0.0f,   // second column
                               0.0f, 0.0f, 1.0f, 0.0f,   // third column
                               0.0f, 0.0f, 0.0f, 1.0f};  // fourth column

//------------------------------------
extern void setup_shaders(const char* vert, const char* frag,
                          GLhandleARB& vs,GLhandleARB& fs, GLhandleARB& prg);
extern void unload_shaders(GLhandleARB& vs, GLhandleARB& fs, GLhandleARB& prg);

//------------------------------------
int load_model (const char* mfile, const char* vfile, const char* ffile) 
{
    ObjModel* m = new ObjModel();
    
    m->obj = obj_add_file(mfile);
    if( m->obj == 0)
    {
        delete m;
        return 0;
    }
    obj_proc_file(m->obj);    

    for( int i = 0; i < 16; i++) 
        m->m[i] = identity[i];
    m->pos[0] = m->pos[1] = m->pos[2] = 0.0f;
    m->scale[0] = m->scale[1] = m->scale[2] = 1.0f;
    setup_shaders( vfile, ffile, m->vs, m->fs, m->prg);
    gModelVec.push_back(m);
    return gModelVec.size() - 1;
}

//------------------------------------
void free_model(int i)
{
    if( i < 0 || i > gModelVec.size() - 1) return;
    if( gModelVec[i] )
    {
        unload_shaders(gModelVec[i]->vs, gModelVec[i]->fs, gModelVec[i]->prg);
        obj_del_file(gModelVec[i]->obj);
        delete gModelVec[i];
        gModelVec[i] = NULL;
    }

    // move all the back ones forward one and pop the back
    for( ; i < gModelVec.size(); i++)
    {
        gModelVec[i] = gModelVec[i + 1];
    }

    gModelVec.pop_back();
}

//------------------------------------
void free_all_models()
{
    for( int i = 0; i < gModelVec.size(); i++)
    {
        unload_shaders(gModelVec[i]->vs, gModelVec[i]->fs, gModelVec[i]->prg);
        obj_del_file(gModelVec[i]->obj);
        delete gModelVec[i];
        gModelVec[i] = NULL;
    }

    gModelVec.clear();
}

//------------------------------------
bool is_model(int i)
{
    if( gModelVec.size() <= 0) return false;
    if( i < 0 || i > gModelVec.size() - 1 ) return false;
    if( gModelVec[i] == NULL ) // fluke
    {
        free_model(i);
        return false;
    }
    
    return true;
}

//------------------------------------
void make_identity(GLfloat *m)
{
    if( !m ) return;
    for(int i = 0; i < 16; i++)
        m[i] = identity[i];
}

//------------------------------------
void copy_matrix(GLfloat *dst, GLfloat* src)
{
    if( !dst || !src ) return;
    for( int i = 0; i < 16; i++)
        dst[i] = src[i];
}

//------------------------------------
// a * b -> d
void mul_matrix(GLfloat *d, GLfloat *a, GLfloat* b)
{

    for( int c = 0; c < 4; c++)
    {
        for( int r = 0; r < 4; r++)
        {
            d[c * 4 + r] = 
                (a[r     ] * b[4 * c    ]) +
                (a[4  + r] * b[4 * c + 1]) +
                (a[8  + r] * b[4 * c + 2]) +
                (a[12 + r] * b[4 * c + 3]) ;
        }
    }
}

//------------------------------------
void print_matrix(GLfloat *m)
{
    printf("Matrix Print Out --\n");
    printf(" %f, %f, %f, %f,\n", m[0], m[4], m[8], m[12]);
    printf(" %f, %f, %f, %f,\n", m[1], m[5], m[9], m[13]);
    printf(" %f, %f, %f, %f,\n", m[2], m[6], m[10], m[14]);
    printf(" %f, %f, %f, %f;\n", m[3], m[7], m[11], m[15]);
}

//------------------------------------
void position_model(int i, float x, float y, float z)
{
    if( !is_model(i) ) return;
        
    gModelVec[i]->pos[0] = x;
    gModelVec[i]->pos[1] = y;
    gModelVec[i]->pos[2] = z;
}

//------------------------------------
void orient_model(int i, float p, float y, float r)
{
    if( !is_model(i) ) return;
    
    GLfloat *m = gModelVec[i]->m;
    GLfloat a[16];
    GLfloat b[16];
    make_identity(m);
    make_identity(a);
    make_identity(b);

    GLfloat cr = cosf(r);
    GLfloat sr = sinf(r);
    GLfloat cp = cosf(p);
    GLfloat sp = sinf(p);
    GLfloat cy = cosf(y);
    GLfloat sy = sinf(y);

    // pitch then yaw then roll
    a[5]  = cp;
    a[6]  = sp;
    a[9]  = -sp;
    a[10] = cp;
    b[0]  = cy;
    b[2]  = -sy;
    b[8]  = sy;
    b[10] = cy;

    mul_matrix(m,a,b);
    copy_matrix(a,m);
    make_identity(b);

    b[0] = cr;
    b[1] = sr;
    b[4] = -sr;
    b[5] = cr;

    mul_matrix(m,a,b);
}

//------------------------------------
void scale_model(int i, float x, float y, float z)
{
    if( !is_model(i) ) return;

    gModelVec[i]->scale[0] = x;
    gModelVec[i]->scale[1] = y;
    gModelVec[i]->scale[2] = z;
}

//------------------------------------
void scale_model(int i, float s)
{
    scale_model(i,s,s,s);
}

//------------------------------------
void translate_model(int i, float x, float y, float z)
{
    if(!is_model(i)) return;

    // orient vector based on our rotation matrix
    float tx, ty, tz;
    GLfloat *m = gModelVec[i]->m;
    tx = (m[0] * x) + (m[4] * y) + (m[8] * z);
    ty = (m[1] * x) + (m[5] * y) + (m[9] * z);
    tz = (m[2] * x) + (m[6] * y) + (m[10] * z);

    // add our new vector to our position vector
    gModelVec[i]->pos[0] += tx;
    gModelVec[i]->pos[1] += ty;
    gModelVec[i]->pos[2] += tz;
}

//------------------------------------
void turn_model(int i, float p, float y, float r)
{
    if(!is_model(i)) return;

    GLfloat *m = gModelVec[i]->m;
    GLfloat a[16];
    GLfloat b[16];
 
    copy_matrix(a,m);
    make_identity(b);

    GLfloat cr = cosf(r);
    GLfloat sr = sinf(r);
    GLfloat cp = cosf(p);
    GLfloat sp = sinf(p);
    GLfloat cy = cosf(y);
    GLfloat sy = sinf(y);

    // pitch then yaw then roll
    b[5]  = cp;
    b[6]  = sp;
    b[9]  = -sp;
    b[10] = cp;
    
//    printf("before transforming\n");
//    print_matrix(m);

    mul_matrix(m,a,b);
    copy_matrix(a,m);
    make_identity(b);
//    printf("after pitch\n");
//    print_matrix(m);

    b[0]  = cy;
    b[2]  = -sy;
    b[8]  = sy;
    b[10] = cy;

    mul_matrix(m,a,b);
    copy_matrix(a,m);
    make_identity(b);
//    printf("after yaw\n");
//    print_matrix(m);

    b[0] = cr;
    b[1] = sr;
    b[4] = -sr;
    b[5] = cr;

    mul_matrix(m,a,b);
//    printf("after roll\n");
//    print_matrix(m);

}

//------------------------------------
void stretch_model(int i, float x, float y, float z)
{
    if(!is_model(i)) return;

    gModelVec[i]->scale[0] *= x;
    gModelVec[i]->scale[1] *= y;
    gModelVec[i]->scale[2] *= z;
}

//------------------------------------
void stretch_model(int i, float s)
{
    stretch_model(i,s,s,s);
}

//------------------------------------
void draw_model(int i)
{
    if(!is_model(i)) return;
    GLfloat *p = gModelVec[i]->pos;
    GLfloat *r = gModelVec[i]->m;
    GLfloat *s = gModelVec[i]->scale;

    // bind the program object
    glUseProgramObjectARB(gModelVec[i]->prg);
    glPushMatrix();
        glTranslatef(-p[0],-p[1],-p[2]);
        glMultMatrixf(r);
        glScalef(s[0],s[1],s[2]);
        obj_draw_file(gModelVec[i]->obj);
    glPopMatrix();

}

//------------------------------------
GLint get_model_uniform_loc(int i, const char* label)
{
    if(!is_model(i)) return -1;
    return glGetUniformLocationARB(gModelVec[i]->prg,label);
}
